# coding=utf8

"""
    rux.models
    ~~~~~~~~~~

    rux's models: blog, author, post, page
"""

from . import src_ext, out_ext, src_dir, out_dir
from .utils import join
from hashlib import md5


class Blog(object):
    """The blog
    attributes
      name          unicode     blog's name
      description   unicode     blog's description
      theme         str         blog's theme"""

    def __init__(self, name="", description="", theme=""):
        self.name = name
        self.description = description
        self.theme = theme


blog = Blog()


class Author(object):
    """The blog's owner, only one
    attributes
      name      unicode     author's name
      email     unicode     author's email
    """

    def __init__(self, name="", email=""):
        self.name = name
        self.email = email

    @property
    def gravatar_id(self):
        """it's md5(author.email), author's gravatar_id"""
        return md5(self.email).hexdigest()


author = Author()


class Post(object):
    """The blog's post object.
    attributes
      name      unicode     post's filename without extension
      title     unicode     post's title
      datetime  datetime    post's created time
      markdown  unicode     post's body source, it's in markdown
      html      unicode     post's html, parsed from markdown
      summary   unicode     post's summary
      filepath  unicode     post's filepath
      title_pic unicode     post's title picture"""

    src_dir = src_dir
    out_dir = join(out_dir, "post")
    template = "post.html"

    def __init__(self, name="", title="", datetime=None, markdown="",
                 html="", summary="", filepath="", title_pic=""):
        self.name = name
        self.title = title
        self.datetime = datetime
        self.markdown = markdown
        self.html = html
        self.summary = summary
        self.filepath = filepath
        self.title_pic = title_pic

    @property
    def src(self):
        return join(Post.src_dir, self.name + src_ext)

    @property
    def out(self):
        return join(Post.out_dir, self.name + out_ext)


class Page(object):
    """The 1st, 2nd, 3rd page..
    attributes
      number    int         the page's order
      posts     list        lists of post objects
      first     bool        is the first page?
      last      bool        is the last page?"""

    template = "page.html"
    out_dir = join(out_dir, "page")

    def __init__(self, number=1, posts=None, first=False, last=False):
        self.number = number
        self.first = first
        self.last = last

        if posts is None:
            self.posts = []
        else:
            self.posts = posts

    @property
    def out(self):
        if self.first:
            return join(out_dir, "index" + out_ext)
        else:
            return join(Page.out_dir, str(self.number) + out_ext)



# coding=utf8

"""
    rux.utils
    ~~~~~~~~~

    All helper functions defined here.
"""

import os
import errno


class Color(object):
    """
    utility to return colored ansi text.
    usage::

        >>> colored("text", "red")
        '\x1b[31mtext\x1b[0m']]'

    """

    colors = {
        'black': 30,
        'red': 31,
        'green': 32,
        'yellow': 33,
        'blue': 34,
        'magenta': 35,
        'cyan': 36,
        'white': 37,
        'bgred': 41,
        'bggrey': 100
    }

    prefix = '\033['
    suffix = '\033[0m'

    def colored(self, text, color=None):

        if color not in self.colors:
            color = 'white'

        clr = self.colors[color]
        return (self.prefix + '%dm%s' + self.suffix) % (clr, text)


colored = Color().colored


def join(*p):
    """return normpath version of path.join"""
    return os.path.normpath(os.path.join(*p))


def chunks(lst, number):
    """
    A generator, split list `lst` into `number` equal size parts.
    usage::

        >>> parts = chunks(range(8),3)
        >>> parts
        <generator object chunks at 0xb73bd964>
        >>> list(parts)
        [[0, 1, 2], [3, 4, 5], [6, 7]]

    """
    lst_len = len(lst)

    for i in xrange(0, lst_len, number):
        yield lst[i: i+number]


def update_nested_dict(a, b):
    """
    update nested dict `a` with another dict b.
    usage::

        >>> a = {'x' : { 'y': 1}}
        >>> b = {'x' : {'z':2, 'y':3}, 'w': 4}
        >>> update_nested_dict(a,b)
        {'x': {'y': 3, 'z': 2}, 'w': 4}

    """
    for k, v in b.iteritems():
        if isinstance(v, dict):
            d = a.setdefault(k, {})
            update_nested_dict(d, v)
        else:
            a[k] = v
    return a


def mkdir_p(path):
    """mkdir -p
    Note: comes from stackoverflow"""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



# coding=utf8

"""
    rux.server
    ~~~~~~~~~~

    rux's server, include a web server and a watcher, running in two threads,
    the file watcher will watch source files updates and start building process
    automatically, the http server host the static site at 0.0.0.0:port
"""

from BaseHTTPServer import BaseHTTPRequestHandler
from BaseHTTPServer import HTTPServer
import logging
from os import listdir as ls
from os import stat
from os.path import exists
from SimpleHTTPServer import SimpleHTTPRequestHandler
import socket
from SocketServer import ThreadingMixIn
import sys
from threading import Thread
from time import sleep

from . import src_ext
from .config import config
from .exceptions import SourceDirectoryNotFound
from .generator import generator
from .logger import logger
from .models import Post
from .utils import join


class Handler(SimpleHTTPRequestHandler):

    def log_message(self, format, *args):
        # http server's output message formatter
        logger.info("%s - %s" % (self.address_string(), format % args))


class MultiThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Multiple threaded http server"""
    pass


class Server(object):
    """rux's server, include a web server to host this static blog at localhost
    , and a files watcher to automatically once source files updated"""

    def __init__(self):
        self.files_stat = {}  # dict, {filepath: file updated time}
        self.server = None  # instance of `MultiThreadedHTTPServer`
        self.watcher = Thread(target=self.watch_files)  # the thread of watcher
        self.watcher.daemon = True  # terminate watcher once main process ends

    def run_server(self, port):
        """run a server binding to port"""

        try:
            self.server = MultiThreadedHTTPServer(('0.0.0.0', port), Handler)
        except socket.error, e:  # failed to bind port
            logger.error(str(e))
            sys.exit(1)

        logger.info("HTTP serve at http://0.0.0.0:%d (ctrl-c to stop) ..."
                    % port)

        try:
            self.server.serve_forever()
        except KeyboardInterrupt:
            logger.info("^C received, shutting down server")
            self.shutdown_server()

    def get_files_stat(self):
        """get source files' update time"""

        if not exists(Post.src_dir):
            logger.error(SourceDirectoryNotFound.__doc__)
            sys.exit(SourceDirectoryNotFound.exit_code)

        paths = []

        for fn in ls(Post.src_dir):
            if fn.endswith(src_ext):
                paths.append(join(Post.src_dir, fn))

        # config.toml
        if exists(config.filepath):
            paths.append(config.filepath)

        # files: a <filepath to updated time> dict
        files = dict((p, stat(p).st_mtime) for p in paths)
        return files

    def watch_files(self):
        """watch files for changes, if changed, rebuild blog. this thread
        will quit if the main process ends"""

        try:
            while 1:
                sleep(1)  # check every 1s

                try:
                    files_stat = self.get_files_stat()
                except SystemExit:
                    logger.error("Error occurred, server shut down")
                    self.shutdown_server()

                if self.files_stat != files_stat:
                    logger.info("Changes detected, start rebuilding..")

                    try:
                        generator.re_generate()
                    except SystemExit:  # catch sys.exit, it means fatal error
                        logger.error("Error occurred, server shut down")
                        self.shutdown_server()

                    self.files_stat = files_stat  # update files' stat
        except KeyboardInterrupt:
            # I dont know why, but this exception won't be catched
            # because absolutly each KeyboardInterrupt is catched by
            # the server thread, which will terminate this thread the same time
            logger.info("^C received, shutting down watcher")
            self.shutdown_watcher()

    def run(self, port):
        """start web server and watcher"""
        self.watcher.start()
        self.run_server(port)

    def shutdown_server(self):
        """shut down the web server"""
        self.server.shutdown()
        self.server.socket.close()

    def shutdown_watcher(self):
        """shut down the watcher thread"""
        self.watcher.join()


server = Server()



# coding=utf8

"""
    rux.parser
    ~~~~~~~~~~

    Parser from post source to html.
"""

from datetime import datetime
import os

from . import charset, src_ext
from .exceptions import *
import libparser

import houdini
import misaka
from misaka import HtmlRenderer, SmartyPants
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter
from pygments.util import ClassNotFound

src_ext_len = len(src_ext)  # cache this, call only once

to_unicode = lambda string: string.decode(charset)


class RuxHtmlRenderer(HtmlRenderer, SmartyPants):
    """misaka render with color codes feature"""

    def _code_no_lexer(self, text):
        # encode to utf8 string
        text = text.encode(charset).strip()
        return(
            """
            <div class="highlight">
              <pre><code>%s</code></pre>
            </div>
            """ % houdini.escape_html(text)
        )

    def block_code(self, text, lang):
        """text: unicode text to render"""

        if not lang:
            return self._code_no_lexer(text)

        try:
            lexer = get_lexer_by_name(lang, stripall=True)
        except ClassNotFound:  # lexer not found, use plain text
            return self._code_no_lexer(text)

        formatter = HtmlFormatter()

        return highlight(text, lexer, formatter)


class Parser(object):
    """Usage::

        parser = Parser()
        parser.parse(str)   # return dict
        parser.markdown.render(markdown_str)  # render markdown to html

    """

    def __init__(self):
        """Initialize the parser, set markdown render handler as
        an attribute `markdown` of the parser"""
        render = RuxHtmlRenderer()  # initialize the color render
        extensions = (
            misaka.EXT_FENCED_CODE |
            misaka.EXT_NO_INTRA_EMPHASIS |
            misaka.EXT_AUTOLINK
        )

        self.markdown = misaka.Markdown(render, extensions=extensions)

    def parse_markdown(self, markdown):
        return self.markdown.render(markdown)

    def parse(self, source):
        """Parse ascii post source, return dict"""

        rt, title, title_pic, markdown = libparser.parse(source)

        if rt == -1:
            raise SeparatorNotFound
        elif rt == -2:
            raise PostTitleNotFound

        # change to unicode
        title, title_pic, markdown = map(to_unicode, (title, title_pic,
                                                      markdown))

        # render to html
        html = self.markdown.render(markdown)
        summary = self.markdown.render(markdown[:200])

        return {
            'title': title,
            'markdown': markdown,
            'html': html,
            'summary': summary,
            'title_pic': title_pic
        }

    def parse_filename(self, filepath):
        """parse post source files name to datetime object"""
        name = os.path.basename(filepath)[:-src_ext_len]
        try:
            dt = datetime.strptime(name, "%Y-%m-%d-%H-%M")
        except ValueError:
            raise PostNameInvalid
        return {'name': name, 'datetime': dt, 'filepath': filepath}


parser = Parser()  # build a runtime parser



# coding=utf8

"""
    rux.logger
    ~~~~~~~~~~

    rux's colorful logger.
"""

from datetime import datetime
import logging
from logging import Formatter
from logging import getLogger
from logging import StreamHandler
import sys

from utils import colored


class ColoredFormatter(Formatter):
    """colored text output formatter"""

    def format(self, record):
        message = record.getMessage()
        mapping = {
            'CRITICAL': 'bgred',
            'ERROR': 'red',
            'WARNING': 'yellow',
            'SUCCESS': 'green',
            'INFO': 'cyan',
            'DEBUG': 'bggrey',
        }
        color = mapping.get(record.levelname, 'white')
        level = colored('%-8s' % record.levelname, color)
        time = colored(datetime.now().strftime("(%H:%M:%S)"), 'magenta')
        return " ".join([level, time, message])


logger = getLogger('rux')
logging.SUCCESS = 25  # WARNING(30) > SUCCESS(25) > INFO(20)
logging.addLevelName(logging.SUCCESS, 'SUCCESS')
logger.success = lambda msg, *args, **kwargs: logger.log(logging.SUCCESS, msg, *args, **kwargs)

# add colored handler
handler = StreamHandler(sys.stdout)
formatter = ColoredFormatter()
handler.setFormatter(formatter)
logger.addHandler(handler)


if __name__ == '__main__':
    """Test all levels out"""
    logger.setLevel(logging.DEBUG)
    logger.info('info')
    logger.success('success')
    logger.debug('debug')
    logger.warning('warning')
    logger.error('error')
    logger.critical('critical')



# coding=utf8

"""
    rux.daemon
    ~~~~~~~~~~

    rux's http server and wacher daemon, it runs rux in the background. This
    daemon is modified from <david@boxedice.com>'s generic daemon class.
"""


import atexit
import logging
import os
import signal
import sys
import time

from .logger import logger
from .server import server


class Daemon(object):

    def __init__(self, pidfile, stdin=os.devnull, stdout=os.devnull,
                 stderr=os.devnull, home_dir='.', umask=022):
        self.pidfile = pidfile
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.home_dir = home_dir
        self.umask = umask
        self.daemon_alive = True

    def daemonize(self, server_port):
        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError, e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno,
                                                            e.strerror))
            sys.exit(1)

        os.chdir(self.home_dir)
        os.setsid()
        os.umask(self.umask)

        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError, e:
            sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno,
                                                            e.strerror))
            sys.exit(1)

        if sys.platform != 'darwin':
            sys.stdout.flush()
            sys.stderr.flush()
            si = file(self.stdin, 'r')
            so = file(self.stdout, 'a+')
            if self.stderr:
                se = file(self.stderr, 'a+', 0)
            else:
                se = so
            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())

        def sigtermhandler(signum, frame):
            self.daemon_alive = False
        signal.signal(signal.SIGTERM, sigtermhandler)
        signal.signal(signal.SIGINT, sigtermhandler)

        logger.success('Started')

        # Write pidfile
        atexit.register(self.delpid)  # Make sure pid file is removed if we \
                                      # quit
        pid = str(os.getpid())
        file(self.pidfile, 'w+').write("%s:%s\n" % (pid, server_port))

    def delpid(self):
        os.remove(self.pidfile)

    def start(self, server_port):
        logger.info('Starting http server(0.0.0.0:%d)'
                    ' and source files watcher..' % server_port)

        try:
            pf = file(self.pidfile, 'r')
            pid, port = map(int, pf.read().strip().split(':'))
            pf.close()
        except (IOError, SystemExit) as e:
            pid, port = None, None

        if pid and port:
            message = ('pidfile %s already exists(pid: %d, port: %d). Is it already running?')
            logger.warning(message % (self.pidfile, pid, port))
            sys.exit(1)

        self.daemonize(server_port)
        self.run(server_port)

    def stop(self):
        logger.info('Stopping the daemon..')

        try:
            pf = file(self.pidfile)
            pid, port = map(int, pf.read().strip().split(':'))
            pf.close()
        except (IOError, ValueError) as e:
            pid, port = None, None

        if not (pid and port):
            message = 'pidfile %s does not exist. Not running?'
            logger.warning(message % self.pidfile)

            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)

            return  # Not an error in a restart process

        # Try killing the daemon process
        try:
            i = 0
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
                i = i + 1
                if i % 10 == 0:
                    os.kill(pid, signal.SIGHUP)
        except OSError, err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                logger.error(str(err))
                sys.exit(1)

        logger.success('Stopped.')

    def status(self):

        try:
            pf = file(self.pidfile)
            pid, port = map(int, pf.read().strip().split(':'))
            pf.close()
        except (IOError, ValueError) as e:
            pid, port = None, None

        if pid and port:
            logger.info('Running: 0.0.0.0:%d, pid: %d.' % (port, pid))
        else:
            logger.info('Stopped.')

    def run(self, port):
        logger.setLevel(logging.ERROR)
        server.run(port)
        logger.setLevel(logging.INFO)


daemon = Daemon("/tmp/rux-daemon.pid", stdout="/dev/stdout")



# coding=utf8

"""
    rux.renderer
    ~~~~~~~~~~~~

    Render data to html with jinja2 templates.
"""

from . import charset
from .exceptions import JinjaTemplateNotFound

from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound


class Renderer(object):

    def initialize(self, templates_path, global_data):
        """initialize with templates' path
        parameters
          templates_path    str    the position of templates directory
          global_data       dict   globa data can be got in any templates"""
        self.env = Environment(loader=FileSystemLoader(templates_path))
        self.env.trim_blocks = True
        self.global_data = global_data

    def render(self, template, **data):
        """Render data with template, return html unicodes.
        parameters
          template   str  the template's filename
          data       dict the data to render
        """
        # make a copy and update the copy
        dct = self.global_data.copy()
        dct.update(data)

        try:
            html = self.env.get_template(template).render(**dct)
        except TemplateNotFound:
            raise JinjaTemplateNotFound
        return html

    def render_to(self, path, template, **data):
        """Render data with template and then write to path"""
        html = self.render(template, **data)
        with open(path, 'w') as f:
            f.write(html.encode(charset))


renderer = Renderer()  # initialized a renderer, and use it each time



# coding=utf8

"""
    rux.generator
    ~~~~~~~~~~~~~

    The core builder processor.
"""

from datetime import datetime
import gc
from os import listdir as ls
from os.path import exists
import sys
import time

from . import src_ext
from .config import config
from .exceptions import *
from .logger import logger
from .models import blog, author, Post, Page
from .parser import parser
from .renderer import renderer
from .utils import chunks, update_nested_dict, mkdir_p, join


def render_to(path, template, **data):
    """shortcut to render data with `template` and then write to `path`.
    Just add exception catch to `renderer.render_to`"""
    try:
        renderer.render_to(path, template, **data)
    except JinjaTemplateNotFound as e:
        logger.error(e.__doc__ + ', Template: %r' % template)
        sys.exit(e.exit_code)


class Generator(object):

    POSTS_COUNT_EACH_PAGE = 15

    def __init__(self):
        self.reset()

    def reset(self):
        self.config = config.default
        self.author = author
        self.blog = blog
        self.posts = []
        self.pages = []

        gc.collect()

    def initialize(self):
        """Initialize configuration and renderer environment"""

        # read configuration
        try:
            conf = config.parse()
        except ConfigSyntaxError as e:
            logger.error(e.__doc__)
            sys.exit(e.exit_code)

        # update default configuration with user defined
        update_nested_dict(self.config, conf)
        self.blog.__dict__.update(self.config['blog'])
        self.author.__dict__.update(self.config['author'])

        # initialize jinja2
        templates = join(self.blog.theme, 'templates')  # templates directory path
        # set a renderer
        jinja2_global_data = {
            'blog': self.blog,
            'author': self.author,
            'config': self.config
        }
        renderer.initialize(templates, jinja2_global_data)
        logger.success('Initialized')

    def get_posts(self):

        if not exists(Post.src_dir):
            logger.error(SourceDirectoryNotFound.__doc__)
            sys.exit(SourceDirectoryNotFound.exit_code)

        source_files = [join(Post.src_dir, fn)
                        for fn in ls(Post.src_dir) if fn.endswith(src_ext)]

        for filepath in source_files:
            try:
                data = parser.parse_filename(filepath)
            except ParseException as e:  # skip single post parse exception
                logger.warn(e.__doc__ + ', filepath: %r' % filepath)
            else:
                self.posts.append(Post(**data))


        # sort posts by its created time, from new to old
        self.posts.sort(key=lambda post: post.datetime.timetuple(), reverse=True)

        count = len(self.posts)

        for idx, post in enumerate(self.posts):

            if idx == 0:
                _next = None
            else:
                _next = self.posts[idx-1]

            if idx == count - 1:
                _prev = None
            else:
                _prev = self.posts[idx+1]

            setattr(post, 'next', _next)
            setattr(post, 'prev', _prev)

    def parse_posts(self):

        for post in self.posts:

            with open(post.filepath, 'r') as file:
                content = file.read()

            try:
                data = parser.parse(content)
            except ParseException, e:
                logger.warn(e.__doc__ + ', filepath %r' % post.filepath)
                pass
            else:
                post.__dict__.update(data)

    def get_pages(self):

        groups = chunks(self.posts, self.POSTS_COUNT_EACH_PAGE)
        self.pages = [Page(number=idx, posts=list(group))
                      for idx, group in enumerate(groups, 1)]

        if self.pages:
            self.pages[0].first = True
            self.pages[-1].last = True

    def render(self):

        mkdir_p(Post.out_dir)
        mkdir_p(Page.out_dir)

        for page in self.pages:
            for post in page.posts:
                render_to(post.out, Post.template, post=post)
            render_to(page.out, Page.template, page=page)

    def generate(self):
        start_time = time.time()
        self.initialize()
        self.get_posts()
        self.get_pages()
        self.parse_posts()
        self.render()

        logger.success("Build done in %.3f seconds" % (time.time() - start_time))

    def re_generate(self):
        self.reset()
        self.generate()


generator = Generator()



# coding=utf8

"""
    rux.libparser
    ~~~~~~~~~~~~~

    Parse post source, return title, title-picture, body(markdown).
"""

import os
from ctypes import *
from distutils.sysconfig import get_python_lib


dll_path = os.path.join(get_python_lib(), 'ruxlibparser.so')

libparser = CDLL(dll_path)


class Post(Structure):
    _fields_ =  (
        ('title', c_void_p),
        ('tpic', c_void_p),
        ('body', c_char_p),
        ('tsz', c_int),
        ('tpsz', c_int)
    )


post = Post()


def parse(src):
    """Note: src should be ascii string"""
    rt = libparser.parse(byref(post), src)
    return (
        rt,
        string_at(post.title, post.tsz),
        string_at(post.tpic, post.tpsz),
        post.body
    )



# coding=utf8

"""
    rux.config
    ~~~~~~~~~~

    Configuration manager, rux's configuration is in toml.
"""

from os.path import exists

from . import charset
from .exceptions import ConfigSyntaxError
from .utils import join

import toml


class Config(object):

    filename = 'config.toml'
    filepath = join('.', filename)

    # default configuration
    default = {
        'blog': {
            'name': '',
            'description': '',
            'theme': 'default',
        },
        'author': {
            'name': 'hit9',
            'email': 'nz2324@126.com',
            'description': 'Who are you?'
        },
        'disqus': {
            'enable': True,
            'shortname': 'rux'
        },
    }

    def parse(self):
        """parse config, return a dict"""

        if exists(self.filepath):
            content = open(self.filepath).read().decode(charset)
        else:
            content = ""

        try:
            config = toml.loads(content)
        except toml.TomlSyntaxError:
            raise ConfigSyntaxError

        return config


config = Config()



# coding=utf8

#     _/_/_/
#    _/    _/  _/    _/  _/    _/
#   _/_/_/    _/    _/    _/_/
#  _/    _/  _/    _/  _/    _/
# _/    _/    _/_/_/  _/    _/
#

"""
    Rux
    ~~~

    Micro & Fast static blog generator (markdown => html).

    :author: Chao Wang (Hit9)
    :license: BSD
"""

__version__ = '0.6.1'

charset = 'utf8'  # utf8 input & output
src_ext = '.md'
out_ext = '.html'
src_dir = 'src'
out_dir = '.'



# coding=utf8

"""
    rux.cli
    ~~~~~~~

    rux's commandline interface.
"""

import sys
import datetime
import logging
from subprocess import call
from os.path import dirname, exists

from . import __version__
from . import src_ext
from .daemon import daemon
from .exceptions import SourceDirectoryNotFound
from .generator import generator
from .pdf import pdf_generator
from .logger import logger
from .models import Post
from .server import server
from .utils import join

from docopt import docopt


usage = """Usage:
  rux [-h|-v]
  rux post
  rux (deploy|build|clean)
  rux (serve|start) [<port>]
  rux (stop|status)
  rux pdf

Options:
  -h --help         show help
  -v --version      show version

Commands:
  post              create a new post
  deploy            create new blog in current directory
  build             build source files to htmls
  serve             start a HTTP server and watch source changes
  clean             remove all htmls rux built
  start             start http server and rebuilder in the background
  stop              stop http server and rebuilder daemon
  status            report the daemon's status
  restart           restart the daemon
  pdf               generate all posts to PDF

Go to http://rux.readthedocs.org/ for more help."""


def deploy_blog():
    """Deploy new blog to current directory"""
    logger.info(deploy_blog.__doc__)
    # `rsync -aqu path/to/res/* .`
    call(
        'rsync -aqu ' + join(dirname(__file__), 'res', '*') + ' .',
        shell=True)
    logger.success('Done')
    logger.info('Please edit config.toml to meet your needs')


def new_post():
    """Touch a new post in src/"""
    logger.info(new_post.__doc__)
    # make the new post's filename
    now = datetime.datetime.now()
    now_s = now.strftime('%Y-%m-%d-%H-%M')
    filepath = join(Post.src_dir, now_s + src_ext)
    # check if `src/` exists
    if not exists(Post.src_dir):
        logger.error(SourceDirectoryNotFound.__doc__)
        sys.exit(SourceDirectoryNotFound.exit_code)
    # write sample content to new post
    content = (
        'Title\n'
        'Title Picture URL\n'
        '---\n'
        'Markdown content ..'
    )
    f = open(filepath, 'w')
    f.write(content)
    f.close()
    logger.success('New post created: %s' % filepath)


def clean():
    """Clean htmls rux built: `rm -rf post page index.html`"""
    logger.info(clean.__doc__)
    paths = ['post', 'page', 'index.html']
    call(['rm', '-rf'] + paths)
    logger.success('Done')


def main():
    arguments = docopt(usage, version=__version__)
    logger.setLevel(logging.INFO)  # !important

    # valiad port argument
    port = arguments['<port>'] or '8888'

    if (not port.isdigit()) or (not 0 < int(port) < 65535):
        logger.error('Port must be an integer in 0-65535.')
        sys.exit(1)
    else:
        port = int(port)

    if arguments['post']:
        new_post()
    elif arguments['deploy']:
        deploy_blog()
    elif arguments['build']:
        generator.generate()
    elif arguments["serve"]:
        server.run(port)
    elif arguments['clean']:
        clean()
    elif arguments['start']:
        daemon.start(port)
    elif arguments['stop']:
        daemon.stop()
    elif arguments['status']:
        daemon.status()
    elif arguments['pdf']:
        pdf_generator.generate()
    else:
        exit(usage)


if __name__ == '__main__':
    main()



# coding=utf8

"""
Rux
---

Micro & Fast static blog generator (markdown => html).

Features
````````

* Static: Markdown => HTML
* Not tags, No categories, No feed generation, No ...
* Minimal & Simple configuration
* Ability to run in the background as a daemon
* Ability to save posts in PDF for offline reading
* Ability to build automatically once source updated

Installation
`````````````

.. code:: bash

    $ mkdir MyBlog
    $ cd MyBlog
    $ virtualenv venv
    New python executable in venv/bin/python
    Installing setuptools............done.
    Installing pip...............done.
    $ . venv/bin/activate
    $ pip install rux

Links
`````

* Documentation <http://rux.readthedocs.org>
* Code on GitHub <https://github.com/hit9/rux>

**NOTICE**: Rux may not be stable before version 1.0

"""

from setuptools import setup, Extension
from rux import __version__


setup(
    name='rux',
    version=__version__,
    author='hit9',
    author_email='nz2324@126.com',
    description='''Micro and fast static blog generator designed only for writing''',
    license='BSD',
    keywords='static blog generator, markdown, html',
    url='http://github.com/hit9/rux',
    packages=['rux'],
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'rux=rux.cli:main'
        ]
    },
    install_requires=open("requirements.txt").read().splitlines(),
    dependency_links=[
        'https://github.com/hit9/toml.py/zipball/master#egg=toml.py-0.1.2',
    ],
    ext_modules=[Extension('ruxlibparser', ['src/libparser.c'])],
    long_description=__doc__,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Customer Service',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
    ]
)



# coding=utf8

"""
    rux.exceptions
    ~~~~~~~~~~~~~~

    All possible exceptions.
"""


class RuxException(Exception):
    """There was an ambiguous exception that occurred while handling
    rux's process"""
    pass


# !Fatal

class RuxFatalError(RuxException):
    """There was a fatal error exception that occurred in rux process"""
    exit_code = 1  # must terminate its process with a non-zero exit code
    pass


class SourceDirectoryNotFound(RuxFatalError):
    """Source directory was not found"""

    exit_code = 2
    pass


class ConfigSyntaxError(RuxFatalError):
    """Toml syntax error occurred in config.toml"""
    exit_code = 3
    pass


class JinjaTemplateNotFound(RuxFatalError):
    """Jinja2 template was not found"""
    exit_code = 4
    pass


# Warning

class RuxWarnException(RuxException):  # warning exception
    """There was a warning exception that occurred in rux process"""
    pass


class ParseException(RuxWarnException):
    """There was an exception while parsing the source"""
    pass


class RenderException(RuxWarnException):
    """There was an exception while rendering to html"""
    pass


class PostNameInvalid(ParseException):
    """Invalid post name, should be datetime, like '1992-04-05-10-10'"""
    # 1992-04-05 is my birthday! :)
    pass


class SeparatorNotFound(ParseException):
    """Separator '---' not found in post source"""
    pass


class PostTitleNotFound(ParseException):
    """There was no title found in post's source"""
    pass



# coding=utf8

"""
    rux.pdf
    ~~~~~~~

    Generate PDF using wkhtmltopdf.
"""

import sys
import time
import subprocess
import os
from os import listdir as ls
from os.path import exists

from . import src_ext, charset
from .config import config
from .exceptions import *
from .parser import parser
from .renderer import renderer
from .logger import logger
from .models import blog, author, Post
from .utils import update_nested_dict, join


def render(template, **data):
    """shortcut to render data with `template`. Just add exception
    catch to `renderer.render`"""
    try:
        return renderer.render(template, **data)
    except JinjaTemplateNotFound as e:
        logger.error(e.__doc__ + ', Template: %r' % template)
        sys.exit(e.exit_code)


class PDFGenerator(object):

    def __init__(self):
        self.commands = ['wkhtmltopdf', '-', 'out.pdf']
        self.config = config.default
        self.blog = blog
        self.author = author
        self.posts = []
        self.html = None

    def initialize(self):

        # read config
        try:
            conf = config.parse()
        except ConfigSyntaxError as e:
            logger.error(e.__doc__)
            sys.exit(e.exit_code)

        update_nested_dict(self.config, conf)

        self.blog.__dict__.update(self.config['blog'])
        self.author.__dict__.update(self.config['author'])

        # initialize jinja2
        templates = join(self.blog.theme, 'templates')
        # set a renderer
        jinja2_global_data = {
            'blog': self.blog,
            'author': self.author,
            'config': self.config
        }
        renderer.initialize(templates, jinja2_global_data)

        logger.success('Initialized')

    def get_posts(self):
        if not exists(Post.src_dir):
            logger.error(SourceDirectoryNotFound.__doc__)
            sys.exit(SourceDirectoryNotFound.exit_code)

        source_files = [join(Post.src_dir, fn)
                        for fn in ls(Post.src_dir) if fn.endswith(src_ext)]

        for filepath in source_files:
            try:
                data = parser.parse_filename(filepath)
            except ParseException as e:  # skip single post parse exception
                logger.warn(e.__doc__ + ', filepath: %r' % filepath)
            else:
                self.posts.append(Post(**data))

        # sort posts by its created time, from new to old
        self.posts.sort(key=lambda post: post.datetime.timetuple(),
                        reverse=True)

    def parse_posts(self):

        for post in self.posts:
            with open(post.filepath, 'r') as file:
                content = file.read()

            try:
                data = parser.parse(content)
            except ParseException, e:
                logger.warn(e.__doc__ + ', filepath %r' % post.filepath)
                pass
            else:
                post.__dict__.update(data)
        logger.success('Posts parsed')

    def render(self):
        self.html = render('pdf.html', posts=self.posts,
                           BLOG_ABS_PATH=os.getcwd())
        logger.success('Posts rendered')

    def generate(self):
        start_time = time.time()
        self.initialize()
        self.get_posts()
        self.parse_posts()
        self.render()

        logger.info('Generate pdf with wkhtmltopdf:')

        try:
            proc = subprocess.Popen(self.commands, stdin=subprocess.PIPE,
                                    stdout=sys.stdout, stderr=sys.stderr)
        except OSError:  # wkhtmltopdf not found
            logger.error('Try to install wkhtmltopdf first %s' %
                         'http://rux.readthedocs.org/en/latest/pdf.html')
            sys.exit(1)
        stdout, stderr = proc.communicate(input=self.html.encode(charset))
        logger.success('Generated to out.pdf in %.3f seconds' %
                       (time.time() - start_time))


pdf_generator = PDFGenerator()
